/*

OpenSource ASE Query SDK 0.1
by Luigi Auriemma
e-mail: aluigi@autistici.org
web:    aluigi.org


INTRODUTION
-----------
This SDK is a simple emulation of the free ASE Query SDK Demo available
on the Udpsoft website and used to put a game server online through the
Udpsoft master servers and to reply to clients queries.
All the informations are in the official original package:

  http://www.udpsoft.com/eye/ASEQuerySDKDemo.zip

Remember that WSADATA must be already initialized in your program.


TODO (probably)
---------------
- the reply status packet is only one so if the data is major than ASE_BUFFSZ it
  will be truncated


LICENSE
-------
    Copyright 2004,2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "ase_v_algo.h"

#ifdef WIN32
    #include <winsock.h>

    #define close   closesocket
    #define sleep   Sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netdb.h>
    #include <fcntl.h>
#endif



    /* ASE defines */

#define ASE_BUFFSZ          2048            /* original is 1364 */
#define ASE_SERVER_1        "master.udpsoft.com"
#define ASE_SERVER_2        "master2.udpsoft.com"
#define ASE_SERVER_PORT     27900
#define ASE_DEF_HBTIME      15              /* initial time between each heartbeat */
#define ASE_DEF_GAMENAME    "test"          /* default gamename used in this SDK */
#define ASE_TEST_ID1        0xd94013c4      /* ID 1 for "test" (ASE_DEF_GAMENAME) */
#define ASE_TEST_ID2        0x5ed6f965      /* ID 2 for "test" */
#define ASE_SERVER_HB       "\\hb\\%hu\\" ASE_DEF_GAMENAME
/*
    The ASE scanner enabled to verify you using the 'v' ping is:
      195.156.109.234   (eye.edome.medialab.sonera.fi)
    Mine is an open SDK so any host can use the 'v' ping
*/

#define ASE_BITPLAYER(x,y)  \
    if(x) { \
        ptr += ase_memcpy(ptr, (char *)x, ase_maxbuff - ptr); \
        *bitnum ^= y; \
    }
#define ASE_DO_HEARTBEAT    \
    if(ase_server1) { \
        ase_peer.sin_addr.s_addr = ase_server1; \
        if(sendto(ase_sd, ase_hb, ase_hblen, 0, (struct sockaddr *)&ase_peer, sizeof(ase_peer)) \
          < 0) return(0); \
    } \
    if(ase_server2) { \
        ase_peer.sin_addr.s_addr = ase_server2; \
        if(sendto(ase_sd, ase_hb, ase_hblen, 0, (struct sockaddr *)&ase_peer, sizeof(ase_peer)) \
          < 0) return(0); \
    }



    /* ASE functions */

int ASEQuery_initialize(
    int     hostport,       /* port to use in the game for game packets             */
    int     internet,       /* 0 = LAN (no master server), 1 = Internet             */
    char    *address        /* IP (char) to bind, NULL = INADDR_ANY                 */
);

int ASEQuery_check(void);

//void ASEQuery_wantstatus(void);

//void ASEQuery_wantrules(void);

//void ASEQuery_wantplayers(void);

void ASEQuery_status(
    const char  *hostname,  /* server name                                          */
    const char  *gametype,  /* type of game                                         */
    const char  *mapname,   /* map name                                             */
    const char  *gamever,   /* game version                                         */
    int         password,   /* 1 = password, 0 = no password                        */
    int         numplayers, /* current players in the server                        */
    int         maxplayers  /* max players                                          */
);

void ASEQuery_addrule(
    const char  *key,       /* new option                                           */
    const char  *value      /* value of the previous option                         */
);

void ASEQuery_addplayer(    /* use 0 to disable a specific player information       */
    const char  *name,
    const char  *team,
    const char  *skin,
    const char  *score,
    const char  *ping,
    const char  *time
);

void ASEQuery_shutdown(void);

int ase_memcpy(
    u_char  *dest,          /* destination string                                   */
    u_char  *src,           /* source string                                        */
    int     maxsz           /* it waste a bit of CPU but is for security reasons!   */
);

u_int ase_resolve(
    char *host              /* hostname or IP to resolve                            */
);



    /* global variables */

u_char  *ase_hb = 0,        /* heartbeat packet                                     */
        *ase_buff = 0,      /* global buffer, useful to gain speed                  */
        *ase_maxbuff,       /* delimits the end of ase_buff (ase_buff + ASE_BUFFSZ  */
        *ASEQuery_error = "ASE error";
u_short ase_port;           /* port assigned for gaming                             */
int     ase_hblen,          /* length of the heartbeat packet                       */
        ase_sd = 0,         /* socket descriptor                                    */
        ase_intlan,         /* 0 = LAN, 1 = Internet                                */
        ase_statuslen;      /* length of ase_status                                 */
u_int  ase_server1,        /* master.udpsoft.com                                   */
        ase_server2,        /* master2.udpsoft.com                                  */
        ase_oldtime,        /* used to know when is the time of a new heartbeat     */
        ase_hbtime;         /* time to wait between each heartbeat, 15, 30, 60, ... */
struct  sockaddr_in ase_peer;



int ASEQuery_initialize(
    int     hostport,
    int     internet,
    char    *address
) {
    struct  sockaddr_in peerb;


    if(ase_sd) return(0);

    ase_server1           = ase_resolve(ASE_SERVER_1);
    ase_server2           = ase_resolve(ASE_SERVER_2);
    if(!ase_server1 && !ase_server2) internet = 0;

    ase_peer.sin_port     = htons(ASE_SERVER_PORT);
    ase_peer.sin_family   = AF_INET;

    if(!address) peerb.sin_addr.s_addr = INADDR_ANY;
        else peerb.sin_addr.s_addr = ase_resolve(address);
    peerb.sin_family      = AF_INET;

    ase_sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(ase_sd < 0) return(0);

    ase_port = hostport;
    hostport += 122;
    do {
        if(!++hostport) return(0);    /* avoid infinite loop on wrong interface */
        peerb.sin_port = htons(hostport);
    } while(bind(ase_sd, (struct sockaddr *)&peerb, sizeof(peerb)));

#ifdef WIN32
    u_int  nonblock = 1;
    if(ioctlsocket(ase_sd, FIONBIO, &nonblock) < 0) return(0);
#else
    if(fcntl(ase_sd, F_SETFL, O_NONBLOCK) < 0) return(0);
#endif

    ase_buff = malloc(ASE_BUFFSZ + 1);
    if(!ase_buff) return(0);
    ase_maxbuff = ase_buff + ASE_BUFFSZ;

    ase_intlan = internet;
    if(!internet) return(1);

    ase_hb = malloc(ASE_BUFFSZ + 1);
    if(!ase_hb) return(0);

    ase_hblen = snprintf(
        ase_hb,
        ASE_BUFFSZ,
        ASE_SERVER_HB,
        hostport);

    ASE_DO_HEARTBEAT;

    ase_oldtime = time(NULL);
    ase_hbtime = ASE_DEF_HBTIME;

    return(1);
}



int ASEQuery_check(void) {
    int     len,
            psz = sizeof(ase_peer);
    u_int  ase_newtime;

    len = recvfrom(ase_sd, ase_buff, ASE_BUFFSZ, 0, (struct sockaddr *)&ase_peer, &psz);
    if(len < 0) {
        if(ase_intlan) {
            ase_newtime = time(NULL);
            if((ase_newtime - ase_oldtime) >= ase_hbtime) {
                ASE_DO_HEARTBEAT;
                ase_oldtime = ase_newtime;
                ase_hbtime <<= 1;
            }
        }
        return(1);
    }

    switch(*ase_buff) {
        case 'p': {             /* PING */
            *ase_buff = 'P';
            } break;
        case 's': {             /* STATUS */
            ASEQuery_wantstatus();
            ASEQuery_wantrules();
            ASEQuery_wantplayers();
            memcpy(ase_buff, "EYE1", 4);    /* read TODO */
            len = ase_statuslen;
            } break;
        case 'g': {             /* GAMENAME */
            memcpy(ase_buff + 1, ASE_DEF_GAMENAME, sizeof(ASE_DEF_GAMENAME) - 1);
            *ase_buff = 'G';
            len = sizeof(ASE_DEF_GAMENAME); /* -1 + 1 */
            } break;
        case 'v': {             /* VERIFY */
            if(len != 9) return(1);
            *(u_int *)(ase_buff + 1) = ase_v_algo(ase_buff + 1, ASE_TEST_ID1, ASE_TEST_ID2);
            *ase_buff = 'V';
            len = 5;
            } break;
        default: return(1);
    }

        /* I think here is better to don't check if < 0 */
    sendto(ase_sd, ase_buff, len, 0, (struct sockaddr *)&ase_peer, psz);
    return(1);
}



void ASEQuery_status(
            const char  *hostname,
            const char  *gametype,
            const char  *mapname,
            const char  *gamever,
            int         password,
            int         numplayers,
            int         maxplayers
) {
    u_char  *ptr = ase_buff,
            num[11];

    ptr += 4;
    ptr += ase_memcpy(ptr, ASE_DEF_GAMENAME, ase_maxbuff - ptr);
    sprintf(num, "%hu", ase_port);  /* a sprintf() here is faster than itoa() */
    ptr += ase_memcpy(ptr, num, ase_maxbuff - ptr);
    ptr += ase_memcpy(ptr, (char *)hostname, ase_maxbuff - ptr);
    ptr += ase_memcpy(ptr, (char *)gametype, ase_maxbuff - ptr);
    ptr += ase_memcpy(ptr, (char *)mapname, ase_maxbuff - ptr);
    ptr += ase_memcpy(ptr, (char *)gamever, ase_maxbuff - ptr);
    sprintf(num, "%d", password);
    ptr += ase_memcpy(ptr, num, ase_maxbuff - ptr);
    sprintf(num, "%d", numplayers);
    ptr += ase_memcpy(ptr, num, ase_maxbuff - ptr);
    sprintf(num, "%d", maxplayers);
    ptr += ase_memcpy(ptr, num, ase_maxbuff - ptr);
    *ptr++ = 0x01;

    ase_statuslen = ptr - ase_buff;
}



void ASEQuery_addrule(
            const char  *key,
            const char  *value
) {
    u_char  *ptr = ase_buff + ase_statuslen - 1;

    ptr += ase_memcpy(ptr, (char *)key, ase_maxbuff - ptr);
    ptr += ase_memcpy(ptr, (char *)value, ase_maxbuff - ptr);
    *ptr++ = 0x01;

    ase_statuslen = ptr - ase_buff;
}



void ASEQuery_addplayer(
            const char  *name,
            const char  *team,
            const char  *skin,
            const char  *score,
            const char  *ping,
            const char  *time
) {
    u_char  *ptr = ase_buff + ase_statuslen,
            *bitnum = ptr++;

    *bitnum = 0;
    ASE_BITPLAYER(name,  1);
    ASE_BITPLAYER(team,  2);
    ASE_BITPLAYER(skin,  4);
    ASE_BITPLAYER(score, 8);
    ASE_BITPLAYER(ping,  16);
    ASE_BITPLAYER(time,  32);

    ase_statuslen = ptr - ase_buff;
}



void ASEQuery_shutdown(void) {
    if(ase_sd)   close(ase_sd);
    if(ase_hb)   free(ase_hb);
    if(ase_buff) free(ase_buff);
}



int ase_memcpy(
            u_char  *dest,
            u_char  *src,
            int     maxsz
) {
    u_char  *beginning = dest++;
    int     i;

/*    if(!src) return(0); useless */

    for(i = 0; i < maxsz; i++) {
        if(!*src) break;
        *dest++ = *src++;
    }
    *beginning = ++i;
    return(i);
}



u_int ase_resolve(char *host) {
    struct hostent *hp;
    u_int host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) return(0);
          else host_ip = *(u_int *)hp->h_addr;
    }
    return(host_ip);
}


